import json
import os
from datetime import datetime, timedelta
from functools import lru_cache
from pathlib import Path

import requests

MAPPINGS_URL = "https://github.com/regro/cf-graph-countyfair/raw/master/mappings/pypi/grayskull_pypi_mapping.yaml"
DOWNLOAD_DIR_ENV_VAR = "PDM_CONDA_PYPI_MAPPING_DIR"


def process_mapping(yaml_path: Path, dict_path: Path):
    """
    Create json mapping from yaml mapping
    :param yaml_path: yaml path
    :param dict_path: json path
    """

    def find_key(key, f):
        while line := f.readline():
            line = line.strip()
            if line.startswith(key):
                return line.split(key)[-1].strip()
        return None

    mappings = dict()
    with yaml_path.open() as f:
        while (conda_name := find_key("conda_name:", f)) is not None:
            pypi_name = find_key("pypi_name:", f)
            if pypi_name:
                mappings[pypi_name] = conda_name

    with dict_path.open("w") as f:
        json.dump(mappings, f)


def download_mapping(download_dir: Path, update_interval: timedelta | None = None) -> dict[str, str]:
    """
    Download and process Conda-PyPI mapping from GitHub
    :param download_dir: download dir
    :param update_interval: update interval, if mapping file modified date is greater than update interval the reload
    :return: Conda mapping
    """
    if update_interval is None:
        update_interval = timedelta(days=15)
    download_dir.mkdir(parents=True, exist_ok=True)
    yaml_path = download_dir / "pypi_mapping.yaml"
    dict_path = yaml_path.with_suffix(".json")

    if not yaml_path.exists() or datetime.fromtimestamp(yaml_path.stat().st_mtime) + update_interval < datetime.now():
        response = requests.get(MAPPINGS_URL, stream=True)
        with yaml_path.open("wb") as f:
            for chunk in response.iter_content(chunk_size=128):
                f.write(chunk)
        process_mapping(yaml_path, dict_path)

    with dict_path.open() as f:
        return json.load(f)


def get_mapping_fixes() -> dict:
    fixes = dict()
    for path in Path(__file__).parents[:3]:
        if (fixes_file := path / "data/mapping_fixes.json").exists():
            with fixes_file.open() as f:
                fixes = json.load(f)
                break
    return fixes


@lru_cache
def get_pypi_mapping() -> dict[str, str]:
    download_dir = os.getenv(DOWNLOAD_DIR_ENV_VAR)
    mapping = download_mapping(Path(str(download_dir)))
    mapping.update(get_mapping_fixes())
    return mapping


@lru_cache
def get_conda_mapping() -> dict[str, str]:
    return {v: k for k, v in get_pypi_mapping().items()}


def _requirement_map(requirement: str, mapping: dict):
    requirement = requirement.strip()
    name = requirement
    for s in (">", "<", "=", "!", "~", " ", "["):
        name = name.split(s, maxsplit=1)[0]
    name = name.split("::")[-1].strip()
    return mapping.get(name, name)


def pypi_to_conda(requirement: str) -> str:
    """
    Map PyPI requirement to Conda version
    :param requirement: PyPI requirement
    :return: Conda requirement
    """
    return _requirement_map(requirement, get_pypi_mapping()).lower()


def conda_to_pypi(requirement: str) -> str:
    """
    Map Conda requirement to PyPI version
    :param requirement: Conda requirement
    :return: PyPI requirement
    """
    return _requirement_map(requirement, get_conda_mapping())



from typing import Callable, Iterator, Mapping, Sequence

from pdm._types import Comparable
from pdm.models.candidates import Candidate
from pdm.models.requirements import Requirement
from pdm.resolver.providers import BaseProvider, EagerUpdateProvider, ReusePinProvider
from pdm.resolver.python import find_python_matches
from resolvelib.resolvers import RequirementInformation
from unearth.utils import LazySequence

from pdm_conda.models.requirements import CondaRequirement


class CondaBaseProvider(BaseProvider):
    def get_preference(
        self,
        identifier: str,
        resolutions: dict[str, Candidate],
        candidates: dict[str, Iterator[Candidate]],
        information: dict[str, Iterator[RequirementInformation]],
        backtrack_causes: Sequence[RequirementInformation],
    ) -> tuple[Comparable, ...]:
        preference = super().get_preference(identifier, resolutions, candidates, information, backtrack_causes)
        return (
            *preference[:3],
            not any(isinstance(i.requirement, CondaRequirement) for i in information[identifier]),
            *preference[3:],
        )

    def find_matches(
        self,
        identifier: str,
        requirements: Mapping[str, Iterator[Requirement]],
        incompatibilities: Mapping[str, Iterator[Candidate]],
    ) -> Callable[[], Iterator[Candidate]]:
        def matches_gen() -> Iterator[Candidate]:
            incompat = list(incompatibilities[identifier])
            if identifier == "python":
                candidates = find_python_matches(identifier, requirements)
                return (c for c in candidates if c not in incompat)
            elif identifier in self.overrides:
                return iter(self.get_override_candidates(identifier))
            reqs = sorted(requirements[identifier], key=self.requirement_preference)
            # iterates over requirements
            candidates = []
            for req in reqs:
                candidates = self._find_candidates(req)
                candidates = LazySequence(
                    can for can in candidates if can not in incompat and all(self.is_satisfied_by(r, can) for r in reqs)
                )
                if candidates:
                    break
            return iter(candidates)

        return matches_gen


class CondaReusePinProvider(ReusePinProvider, CondaBaseProvider):
    pass


class CondaEagerUpdateProvider(EagerUpdateProvider, CondaBaseProvider):
    pass



import functools

from pdm.models.candidates import Candidate
from pdm.models.requirements import Requirement
from pdm.project import Project

from pdm_conda.models.candidates import CondaCandidate
from pdm_conda.models.requirements import CondaRequirement

_patched = False


def remove_quotes(req: str) -> str:
    for quote in ("'", '"'):
        if req.startswith(quote) and req.endswith(quote):
            req = req[1:-1]
    return req


def wrap_save_version_specifiers(func):
    @functools.wraps(func)
    def wrapper(
        requirements: dict[str, dict[str, Requirement]],
        resolved: dict[str, Candidate],
        save_strategy: str,
    ) -> None:
        func(requirements, resolved, save_strategy)
        for reqs in requirements.values():
            for name, r in reqs.items():
                if isinstance(r, CondaRequirement):
                    r.version_mapping.update(resolved[name].req.version_mapping)
                    r.is_python_package = resolved[name].req.is_python_package

    return wrapper


def wrap_format_lockfile(func):
    @functools.wraps(func)
    def wrapper(
        project: Project,
        mapping: dict[str, Candidate],
        fetched_dependencies: dict[tuple[str, str | None], list[Requirement]],
    ) -> dict:
        for deps in fetched_dependencies.values():
            for i, dep in enumerate(deps):
                if isinstance(dep, CondaRequirement):
                    setattr(dep, "as_line", functools.partial(dep.as_line, with_build_string=True))
        version_mapping = {}
        for name, can in mapping.items():
            if isinstance(can, CondaCandidate):
                fetched_dependencies[(can.name, can.conda_version)] = fetched_dependencies.pop(
                    (can.name, can.version),
                    [],
                )
                version_mapping[name] = can.version
                can.version = can.conda_version
        try:
            res = func(project, mapping, fetched_dependencies)
            return res
        finally:
            for deps in fetched_dependencies.values():
                for i, dep in enumerate(deps):
                    if isinstance(dep, CondaRequirement):
                        setattr(dep, "as_line", dep.as_line.func)  # type: ignore
            for name, version in version_mapping.items():
                can = mapping[name]
                can.version = version
                fetched_dependencies[(can.name, can.version)] = fetched_dependencies.pop(
                    (can.name, can.conda_version),
                    [],
                )

    return wrapper


if not _patched:
    from pdm.cli import actions, utils

    save_version_specifiers = wrap_save_version_specifiers(utils.save_version_specifiers)
    format_lockfile = wrap_format_lockfile(utils.format_lockfile)
    for m in [utils, actions]:
        setattr(m, "save_version_specifiers", save_version_specifiers)
        setattr(m, "format_lockfile", format_lockfile)
    _patched = True



import uuid
from copy import copy
from typing import Any, Iterable, Mapping, cast

from pdm._types import RepositoryConfig
from pdm.models.repositories import BaseRepository, LockedRepository, PyPIRepository
from pdm.models.requirements import Requirement
from pdm.models.specifiers import PySpecSet
from pdm.resolver.python import PythonRequirement

from pdm_conda.conda import conda_create, sort_candidates
from pdm_conda.environments import BaseEnvironment, CondaEnvironment
from pdm_conda.models.candidates import Candidate, CondaCandidate
from pdm_conda.models.requirements import (
    CondaRequirement,
    NamedRequirement,
    as_conda_requirement,
)


class CondaRepository(BaseRepository):
    def __init__(
        self,
        sources: list[RepositoryConfig],
        environment: BaseEnvironment,
        ignore_compatibility: bool = True,
    ) -> None:
        super().__init__(sources, environment, ignore_compatibility)
        self.environment = cast(CondaEnvironment, environment)
        self._conda_resolution: dict[str, list[CondaCandidate]] = dict()

    def _uses_conda(self, requirement: Requirement) -> bool:
        """
        True if requirement is conda requirement or (not excluded and named requirement
        and conda as default manager or used by another conda requirement)
        :param requirement: requirement to evaluate
        """
        if not isinstance(self.environment, CondaEnvironment):
            return False
        conda_config = self.environment.project.conda_config
        return isinstance(requirement, (CondaRequirement, PythonRequirement)) or (
            isinstance(requirement, NamedRequirement)
            and conda_config.as_default_manager
            and requirement.name not in conda_config.excludes
        )

    def update_conda_resolution(
        self,
        requirements: list[Requirement] | None = None,
        resolution: dict | None = None,
    ) -> list[CondaRequirement]:
        """
        Updates the existing conda resolution if new requirements.
        :param requirements: list of requirements
        :param resolution: resolution to override existing
        :return: list of changed requirements
        """
        if resolution is not None:
            self._conda_resolution = resolution
        return []

    def get_dependencies(self, candidate: Candidate) -> tuple[list[Requirement], PySpecSet, str]:
        if isinstance(candidate, CondaCandidate):
            dependencies = list(candidate.dependencies)
            # if dep in constrains use constrain
            if candidate.constrains:
                for i, dep in enumerate(dependencies):
                    if (constrain := candidate.constrains.get(dep.identify(), None)) is not None:
                        dependencies[i] = constrain
            requires_python = PySpecSet(candidate.requires_python)
            summary = candidate.summary
        else:
            dependencies, requires_python, summary = super().get_dependencies(candidate)
        return dependencies, requires_python, summary


class PyPICondaRepository(PyPIRepository, CondaRepository):
    def update_conda_resolution(
        self,
        requirements: list[Requirement] | None = None,
        resolution: dict | None = None,
    ) -> list[CondaRequirement]:
        super().update_conda_resolution(requirements, resolution)
        changed = []
        requirements = requirements or []
        requirements = [as_conda_requirement(req) for req in requirements if self._uses_conda(req)]
        update = False
        # if any requirement not in saved candidates or incompatible candidates
        for req in requirements:
            if update:
                break
            key = req.identify()
            if key not in self._conda_resolution:
                update = True
                break
            for can in self._conda_resolution[key]:
                if not req.is_compatible(can):
                    update = True
                    break

        if update:
            resolution = conda_create(
                self.environment.project,
                requirements,
                prefix=f"/tmp/{uuid.uuid4()}",
                dry_run=True,
            )
            _requirements = {r.conda_name: r for r in requirements}
            for name, candidates in resolution.items():
                req = _requirements.get(name, candidates[0].req)
                key = req.identify()
                cans = self._conda_resolution.get(key, [])
                # add non-existing candidates
                if any(True for can in candidates if can not in cans):
                    # add requirement to changed if didn't exist
                    if key in self._conda_resolution:
                        changed.append(req)
                    self._conda_resolution[key] = candidates

        return changed

    def _find_candidates(self, requirement: Requirement) -> Iterable[Candidate]:
        if self._uses_conda(requirement):
            requirement = as_conda_requirement(requirement)
            candidates = self._conda_resolution.get(requirement.identify(), [])
            candidates = [copy(c) for c in candidates if requirement.is_compatible(c)]
            candidates = list(sort_candidates(self.environment.project, candidates))
            for can in candidates:
                can.req = requirement
        else:
            candidates = super()._find_candidates(requirement)
        return candidates


class LockedCondaRepository(LockedRepository, CondaRepository):
    def _read_lockfile(self, lockfile: Mapping[str, Any]) -> None:
        packages = lockfile.get("package", [])
        conda_packages = [copy(p) for p in packages if p.get("conda_managed", False)]
        packages = [p for p in packages if not p.get("conda_managed", False)]
        super()._read_lockfile({"package": packages, "metadata": lockfile.get("metadata", {})})

        for package in conda_packages:
            link, _hash = list(self.file_hashes[(package["name"], package["version"])].items())[0]
            name, value = _hash.split(":", maxsplit=1)
            package[name] = value
            package["url"] = link.url_without_fragment
            can = CondaCandidate.from_lock_package(package)
            can_id = self._identify_candidate(can)
            self.packages[can_id] = can
            self.candidate_info[can_id] = (
                can.dependencies_lines,
                package.get("requires_python", ""),
                package.get("summary", ""),
            )

    def _identify_candidate(self, candidate: Candidate) -> tuple:
        if isinstance(candidate, CondaCandidate):
            return candidate.identify(), candidate.version, None, False
        return super()._identify_candidate(candidate)



import hashlib
import json

from pdm.project.project_file import PyProject as PyProjectBase


class PyProject(PyProjectBase):
    def content_hash(self, algo: str = "sha256") -> str:
        """
        Generate a hash of the sensible content of the pyproject.toml file.
        When the hash changes, it means the project needs to be relocked.
        :param algo: hash algorithm name
        :return: pyproject.toml hash
        """
        pdm_conda_data = self.settings.get("conda", {})
        dump_data = {
            "sources": self.settings.get("source", []),
            "dependencies": self.metadata.get("dependencies", []),
            "dev-dependencies": self.settings.get("dev-dependencies", {}),
            "optional-dependencies": self.metadata.get("optional-dependencies", {}),
            "requires-python": self.metadata.get("requires-python", ""),
            "pdm-conda": {
                "channels": pdm_conda_data.get("channels", []),
                "as-default-manager": pdm_conda_data.get("as-default-manager", False),
                "excludes": pdm_conda_data.get("excludes", []),
                "dependencies": pdm_conda_data.get("dependencies", []),
                "dev-dependencies": pdm_conda_data.get("dev-dependencies", {}),
                "optional-dependencies": pdm_conda_data.get("optional-dependencies", {}),
            },
            "overrides": self.resolution_overrides,
        }
        pyproject_content = json.dumps(dump_data, sort_keys=True)
        hasher = hashlib.new(algo)
        hasher.update(pyproject_content.encode("utf-8"))
        return hasher.hexdigest()



import os
from contextlib import contextmanager
from dataclasses import dataclass, field
from enum import Enum
from functools import wraps
from pathlib import Path
from typing import Any

from pdm.exceptions import ProjectError
from pdm.project import ConfigItem, Project

from pdm_conda.mapping import DOWNLOAD_DIR_ENV_VAR


class CondaRunner(str, Enum):
    CONDA = "conda"
    MAMBA = "mamba"
    MICROMAMBA = "micromamba"


class CondaSolver(str, Enum):
    CONDA = "conda"
    MAMBA = "libmamba"


_CONFIG_MAP = {"pypi-mapping.download-dir": "mapping_download_dir"}
_CONFIG_MAP |= {v: k for k, v in _CONFIG_MAP.items()}

CONFIGS = [
    ("runner", ConfigItem("Conda runner executable", CondaRunner.CONDA.value, env_var="PDM_CONDA_RUNNER")),
    ("solver", ConfigItem("Solver to use for Conda resolution", CondaSolver.CONDA.value, env_var="PDM_CONDA_SOLVER")),
    ("channels", ConfigItem("Conda channels to use")),
    (
        "as-default-manager",
        ConfigItem("Use Conda to install all possible requirements", False, env_var="PDM_CONDA_AS_DEFAULT_MANAGER"),
    ),
    (
        "batched-commands",
        ConfigItem("Execute batched install and remove commands", False, env_var="PDM_CONDA_BATCHED_COMMANDS"),
    ),
    (
        "installation-method",
        ConfigItem(
            "Whether to use hard-link or copy when installing",
            "hard-link",
            env_var="PDM_CONDA_INSTALLATION_METHOD",
        ),
    ),
    ("dependencies", ConfigItem("Dependencies to install with Conda")),
    ("optional-dependencies", ConfigItem("Optional dependencies to install with Conda")),
    ("dev-dependencies", ConfigItem("Development dependencies to install with Conda")),
    ("excludes", ConfigItem("Excluded dependencies from Conda")),
    (
        "pypi-mapping.download-dir",
        ConfigItem(
            "PyPI-Conda mapping download directory",
            Path().home() / ".pdm-conda/",
            env_var=DOWNLOAD_DIR_ENV_VAR,
        ),
    ),
]

CONFIGS = [(f"conda.{name}", config) for name, config in CONFIGS]


def is_decorated(func):
    return hasattr(func, "__wrapped__")


def is_conda_config_initialized(project: Project):
    return "conda" in project.pyproject.settings


@dataclass
class PluginConfig:
    _project: Project = field(repr=False, default=None)
    _initialized: bool = field(repr=False, default=False, compare=False)
    _set_project_config: bool = field(repr=False, default=False, compare=False)

    channels: list[str] = field(default_factory=list)
    runner: str = CondaRunner.CONDA
    solver: str = CondaSolver.CONDA
    as_default_manager: bool = False
    batched_commands: bool = False
    installation_method: str = "hard-link"
    excludes: list[str] = field(default_factory=list, repr=False)
    dependencies: list[str] = field(default_factory=list, repr=False)
    optional_dependencies: dict[str, list] = field(default_factory=dict)
    dev_dependencies: dict[str, list] = field(default_factory=dict)
    mapping_download_dir: Path = field(repr=False, default=Path())

    def __post_init__(self):
        if self.runner not in list(CondaRunner):
            raise ProjectError(f"Invalid Conda runner: {self.runner}")
        if self.solver not in list(CondaSolver):
            raise ProjectError(f"Invalid Conda solver: {self.solver}")
        if self.installation_method not in ["hard-link", "copy"]:
            raise ProjectError(f"Invalid Conda installation method: {self.installation_method}")
        to_suscribe = [
            (self._project.pyproject._data, "update"),
            (self._project.pyproject, "write"),
            (self._project.pyproject, "reload"),
        ]
        for obj, name in to_suscribe:
            func = getattr(obj, name)
            if not is_decorated(func):
                setattr(obj, name, self.suscribe(self, func))
        self._set_project_config = True
        if self.is_initialized:
            config = self._project.project_config
            config["python.use_venv"] = True
            config["python.use_pyenv"] = False
            config["venv.backend"] = self.runner
            if (venv_location := os.getenv("VIRTUAL_ENV", os.getenv("CONDA_PREFIX", None))) is not None:
                self._project.global_config["venv.location"] = str(Path(venv_location).parent)
                os.environ["PDM_IGNORE_SAVED_PYTHON"] = "true"
                os.environ.pop("PDM_IGNORE_ACTIVE_VENV", None)

    def __setattr__(self, name: str, value: Any) -> None:
        super().__setattr__(name, value)
        # if plugin config is set then maybe update pyproject settings
        if not name.startswith("_") and not callable(getattr(self, name)):
            name = f"conda.{_CONFIG_MAP.get(name, name)}".replace("_", "-")
            name, config_item = next(filter(lambda n: name == n[0], CONFIGS))
            if self._set_project_config:
                name_path = name.split(".")
                name = name_path.pop(-1)
                config = self._project.pyproject.settings
                for p in name_path:
                    config = config.setdefault(p, dict())
                config[name] = value
                self._initialized = True
            if config_item.env_var:
                os.environ.setdefault(config_item.env_var, str(value))

    def reload(self):
        _conf = self.load_config(self._project)
        with self.omit_set_project_config():
            for k, v in _conf.__dict__.items():
                if not callable(v) and k not in ("_project", "_set_project_config") and getattr(self, k) != v:
                    setattr(self, k, v)

    @staticmethod
    def suscribe(config, func):
        """
        Suscribe to function call and after executed refresh config
        :param config: PluginConfig to refresh
        :param func: function to decorate
        """

        @wraps(func)
        def wrapper(*args, **kwargs):
            result = func(*args, **kwargs)
            config.reload()
            return result

        return wrapper

    @contextmanager
    def with_config(self, **kwargs):
        """
        Context manager that temporarily updates configs without updating pyproject settings
        """
        configs = ["_set_project_config"] + list(kwargs)
        kwargs["_set_project_config"] = False
        old_values = {}
        for name in configs:
            old_values[name] = getattr(self, name)
            setattr(self, name, kwargs[name])
        try:
            yield
        finally:
            for name in reversed(configs):
                setattr(self, name, old_values[name])

    @contextmanager
    def omit_set_project_config(self):
        """
        Context manager that deactivates updating pyproject settings
        """
        with self.with_config():
            yield

    @property
    def is_initialized(self):
        return self._initialized

    @classmethod
    def load_config(cls, project: Project, **kwargs) -> "PluginConfig":
        """
        Load plugin configs from project settings.
        :param project: Project
        :param kwargs: settings overwrites
        :return: plugin configs
        """
        config = {k: v for k, v in project.pyproject.settings.get("conda", {}).items()}
        kwargs["_initialized"] = is_conda_config_initialized(project)
        for n, c in CONFIGS:
            n = n[len("conda.") :]
            if (prop_name := _CONFIG_MAP.get(n, n)) not in config and c.env_var:
                value = project.config[f"conda.{n}"]
                if prop_name == "mapping_download_dir":
                    value = Path(value)
                elif prop_name in ("as-default-manager", "batched-commands"):
                    value = str(value).lower() in ("true", "1")
                config[prop_name] = value
        config = {k.replace("-", "_"): v for k, v in config.items()}
        return PluginConfig(_project=project, **(config | kwargs))

    def command(self, cmd="install"):
        """
        Get runner command args
        :param cmd: command, install by default
        :return: args list
        """
        runner = self.runner
        if cmd == "remove" and runner == CondaRunner.MAMBA:
            runner = CondaRunner.CONDA
        if isinstance(runner, CondaRunner):
            runner = runner.value

        _command = [runner, *cmd.split(" ")]

        if self.runner != CondaRunner.CONDA and cmd == "search":
            _command.insert(1, "repoquery")
        if cmd in ("install", "remove", "create"):
            _command.append("-y")
        if cmd in ("install", "create") or (cmd == "search" and self.runner == CondaRunner.MICROMAMBA):
            _command.append("--strict-channel-priority")
        if self.runner == CondaRunner.CONDA and self.solver == CondaSolver.MAMBA and cmd in ("create", "install"):
            _command.extend(["--solver", CondaSolver.MAMBA.value])
        return _command



import re
from importlib.metadata import Distribution
from pathlib import Path
from typing import Any, cast
from urllib.parse import urlparse

from pdm.environments import BaseEnvironment
from pdm.models.candidates import Candidate, PreparedCandidate
from pdm.models.setup import Setup
from unearth import Link

from pdm_conda.models.requirements import (
    CondaRequirement,
    Requirement,
    as_conda_requirement,
    parse_conda_version,
    parse_requirement,
)
from pdm_conda.models.setup import CondaSetupDistribution


def parse_channel(channel_url: str) -> str:
    """
    Parse channel from channel url
    :param channel_url: channel url from package
    :return: channel
    """
    channel = urlparse(channel_url).path
    if channel.startswith("/"):
        channel = channel[1:]
    return channel


class CondaPreparedCandidate(PreparedCandidate):
    def __init__(self, candidate: Candidate, environment: BaseEnvironment) -> None:
        super().__init__(candidate, environment)
        self.candidate = cast(CondaCandidate, self.candidate)  # type: ignore

    def get_dependencies_from_metadata(self) -> list[str]:
        # if conda candidate return already obtained dependencies
        return [d.as_line(as_conda=True, with_build_string=True) for d in self.candidate.dependencies]

    def prepare_metadata(self) -> Distribution:
        # if conda candidate get setup from package
        return self.candidate.distribution


class CondaCandidate(Candidate):
    def __init__(
        self,
        req: Requirement,
        name: str | None = None,
        version: str | None = None,
        link: Link | None = None,
        dependencies: list[str] | None = None,
        constrains: list[str] | None = None,
        build_string: str | None = None,
        build_number: int = 0,
        timestamp: int = 0,
        channel: str | None = None,
        track_feature: str = "",
    ):
        super().__init__(req, name, version, link)
        self._req = cast(CondaRequirement, req)  # type: ignore
        self._preferred = None
        self._prepared: CondaPreparedCandidate | None = None
        self.dependencies: list[CondaRequirement] = [
            cast(CondaRequirement, parse_requirement(f"conda:{r}")) for r in (dependencies or [])
        ]
        self.constrains: dict[str, CondaRequirement] = dict()
        self.hashes = {self.link: f"{self.link.hash_name}:{self.link.hash}"}
        for r in constrains or []:
            c = cast(CondaRequirement, parse_requirement(f"conda:{r}"))
            self.constrains[str(c.conda_name)] = c
        self.build_string = build_string
        self.build_number = build_number
        self.timestamp = timestamp
        self.channel = channel
        self.track_feature = track_feature
        self.conda_version = version
        self.version = parse_conda_version(version)

    @property
    def req(self):
        return self._req

    @req.setter
    def req(self, value):
        if isinstance(value, CondaRequirement):
            self._req = value

    @property
    def distribution(self):
        return CondaSetupDistribution(
            Setup(
                name=self.name,
                summary="",
                version=self.version,
                install_requires=self.dependencies_lines,
                python_requires=self.requires_python,
            ),
        )

    @property
    def dependencies_lines(self):
        return [dep.as_line(as_conda=True, with_build_string=True, with_channel=True) for dep in self.dependencies]

    def as_lockfile_entry(self, project_root: Path) -> dict[str, Any]:
        result = super().as_lockfile_entry(project_root)
        result["conda_managed"] = True
        if self.link is None:
            raise ValueError("Uninitialized conda requirement")
        result["channel"] = self.channel
        if self.build_string is not None:
            result["build_string"] = self.build_string
        result["build_number"] = self.build_number
        if self.track_feature:
            result["track_feature"] = self.track_feature
        if self.constrains:
            result["constrains"] = [c.as_line(with_build_string=True) for c in self.constrains.values()]
        result["version"] = self.conda_version
        return result

    def prepare(self, environment: BaseEnvironment) -> CondaPreparedCandidate:
        """Prepare the candidate for installation."""
        if self._prepared is None:
            self._prepared = CondaPreparedCandidate(self, environment)
        return self._prepared

    @classmethod
    def from_lock_package(cls, package: dict) -> "CondaCandidate":
        """
        Create conda candidate from lockfile package.
        :param package: lockfile package
        :return: conda candidate
        """
        requires_python = package.get("requires_python", "")
        dependencies = package.get("dependencies", [])
        if requires_python:
            dependencies.append(f"python {requires_python}")
        return CondaCandidate.from_conda_package(package | {"depends": dependencies})

    @classmethod
    def from_conda_package(cls, package: dict, requirement: CondaRequirement | None = None) -> "CondaCandidate":
        """
        Create conda candidate from conda package.
        :param package: conda package
        :param requirement: conda requirement associated with conda package
        :return: conda candidate
        """
        dependencies: list = package["depends"] or []
        requires_python = None
        to_delete = []
        for d in dependencies:
            if d.startswith("__"):
                to_delete.append(d)
            elif match := re.match(r"python( .+|$)", d):
                to_delete.append(d)
                if requires_python is None:
                    requires_python = match.group(1).strip().split(" ")[0] or "*"
        for d in to_delete:
            dependencies.remove(d)
        hashes = {h: package[h] for h in ["md5"] if h in package}
        url = package["url"]
        for k, v in hashes.items():
            url += f"#{k}={v}"
        name, version = package["name"], package["version"]
        build_string = package.get("build", package.get("build_string", ""))
        channel = parse_channel(package["channel"])
        if requirement is not None:
            requirement = cast(CondaRequirement, as_conda_requirement(requirement))
            requirement.version_mapping.update({parse_conda_version(version): version})
        else:
            requirement = parse_requirement(f"conda:{name} {version} {build_string}")

        assert requirement is not None
        requirement.is_python_package = requires_python is not None
        return CondaCandidate(
            req=requirement,
            name=name,
            version=version,
            link=Link(
                url,
                requires_python=requires_python,
                comes_from=channel,
            ),
            channel=channel,
            dependencies=dependencies,
            constrains=package.get("constrains", None) or [],
            build_string=build_string,
            build_number=package.get("build_number", 0),
            track_feature=package.get("track_feature", ""),
            timestamp=package.get("timestamp", 0),
        )

    def __str__(self) -> str:
        if self.req.is_named:
            return f"{self.name}@{self.conda_version}"
        return super().__str__()

    def format(self) -> str:
        """Format for output."""
        return f"[req]{self.name}[/] [warning]{self.conda_version}[/]"



import re

_patched = False


def normalize_name(name: str, lowercase: bool = True) -> str:
    """
    Normalize name and keep _
    :param name: name to normalize
    :param lowercase: if true lowercase name
    :return: normalized name
    """
    name = re.sub(r"[^A-Za-z0-9._]+", "-", name)
    return name.lower() if lowercase else name


if not _patched:
    from pdm.cli import actions, utils
    from pdm.installers import synchronizers
    from pdm.models import candidates, repositories, requirements, working_set

    modules = [
        actions,
        utils,
        synchronizers,
        candidates,
        requirements,
        repositories,
        working_set,
    ]
    for m in modules:
        setattr(m, "normalize_name", normalize_name)

    _patched = True



from typing import Any, cast

from pdm.models.setup import Setup, SetupDistribution

from pdm_conda.models.requirements import CondaRequirement, parse_requirement


class CondaSetupDistribution(SetupDistribution):
    def __init__(self, data: Setup, package: dict | None = None) -> None:
        self.package = package or dict()
        super().__init__(data)

    @property
    def metadata(self) -> dict[str, Any]:
        metadata = super().metadata
        metadata["Home-Page"] = None
        metadata["License"] = "other"
        return metadata

    def as_line(self):
        channel = self.package.get("channel", "")
        if channel:
            channel += "::"
        build_string = self.package.get("build_string", self.package.get("build", ""))
        if build_string:
            build_string = f" {build_string}"
        version = self.package.get("version", "")
        name = self.package.get("name", "")
        return f"{channel}{name}=={version}{build_string}"

    @property
    def req(self) -> CondaRequirement:
        return cast(CondaRequirement, parse_requirement(f"conda:{self.as_line()}"))



import os
import sysconfig
import uuid
from pathlib import Path
from typing import cast

from pdm.environments import PythonEnvironment
from pdm.exceptions import ProjectError
from pdm.models.requirements import Requirement
from pdm.models.specifiers import PySpecSet
from pdm.models.working_set import WorkingSet
from pdm.project import Project

from pdm_conda.conda import conda_create, conda_list, conda_search
from pdm_conda.mapping import pypi_to_conda
from pdm_conda.models.config import CondaRunner, CondaSolver
from pdm_conda.project import CondaProject
from pdm_conda.utils import normalize_name


def ensure_conda_env():
    if (packages_path := os.getenv("CONDA_PREFIX", None)) is None:
        raise ProjectError("Conda environment not detected.")
    return packages_path


class CondaEnvironment(PythonEnvironment):
    def __init__(self, project: Project) -> None:
        super().__init__(project)
        self.project = cast(CondaProject, project)
        self._env_dependencies: dict[str, Requirement] | None = None
        self.python_requires &= PySpecSet(f"=={self.interpreter.version}")

    @property
    def packages_path(self) -> Path:
        return Path(ensure_conda_env())

    def get_paths(self) -> dict[str, str]:
        prefix = ensure_conda_env()
        paths = sysconfig.get_paths(vars={k: prefix for k in ("base", "platbase", "installed_base")}, expand=True)
        paths.setdefault("prefix", prefix)
        return paths

    def get_working_set(self) -> WorkingSet:
        """
        Get the working set based on local packages directory, include Conda managed packages.
        """
        working_set = super().get_working_set()
        working_set._dist_map = conda_list(self.project) | {
            normalize_name(pypi_to_conda(dist.metadata["Name"])): dist for dist in working_set._dist_map.values()
        }
        return working_set

    @property
    def env_dependencies(self) -> dict[str, Requirement]:
        if self._env_dependencies is None:
            self._env_dependencies = dict()

            def load_dependencies(name: str, packages: dict, dependencies: dict):
                if name not in packages and name not in dependencies:
                    return
                candidate = conda_search(self.project, packages[name].req)[0]
                dependencies[name] = candidate.req
                for d in candidate.dependencies:
                    load_dependencies(d.name, packages, dependencies)

            working_set = conda_list(self.project)
            dependencies = ["python"]
            if (runner := self.project.conda_config.runner) in working_set:
                dependencies.append(runner)
            if (
                runner in (CondaRunner.MAMBA, CondaRunner.MICROMAMBA)
                or self.project.conda_config.solver == CondaSolver.MAMBA
            ):
                self._env_dependencies = conda_create(
                    self.project,
                    [working_set[d].req for d in dependencies],
                    prefix=f"/tmp/{uuid.uuid4()}",
                    dry_run=True,
                )
            else:
                for dep in dependencies:
                    load_dependencies(dep, working_set, self._env_dependencies)

        return self._env_dependencies



import argparse
from typing import cast

from pdm.cli.commands.lock import Command as BaseCommand

from pdm_conda.project import CondaProject, Project


class Command(BaseCommand):
    description = BaseCommand.__doc__
    name = "lock"

    def handle(self, project: Project, options: argparse.Namespace) -> None:
        project = cast(CondaProject, project)
        if options.groups:
            if ":all" in options.groups:
                options.groups += list(project.iter_groups())
        super().handle(project, options)
